@@ -1,7 +1,7 @@
 # BEGIN BPS TAGGED BLOCK {{{
 # COPYRIGHT:
 # 
-# This software is Copyright (c) 2003-2006 Best Practical Solutions, LLC
+# This software is Copyright (c) 2003-2008 Best Practical Solutions, LLC
 #                                          <clkao@bestpractical.com>
 # 
 # (Except where explicitly superseded by other copyright notices)
@@ -68,7 +68,9 @@ our @EXPORT_OK = qw(
     move_path make_path splitpath splitdir tmpdir tmpfile get_depot_anchor
     catdepot abs_path_noexist 
 
-    is_symlink is_executable is_uri can_run is_path_inside
+    is_symlink is_executable is_uri can_run is_path_inside is_depotpath
+
+    uri_escape uri_unescape
 
     str2time time2str reformat_svn_date
 
@@ -193,7 +195,7 @@ sub get_prompt { {
     my $out = (IS_WIN32 ? sub { 1 } : sub { print @_ });
 
     my $erase;
-    if (!IS_WIN32) {
+    if (!IS_WIN32 && -t) {
        my %keys = Term::ReadKey::GetControlChars();
        $erase = $keys{ERASE};
     }
@@ -812,6 +814,7 @@ sub can_run {
 
     for my $dir ((split /$Config::Config{path_sep}/, $ENV{PATH}), @path, '.') {
         my $abs = catfile($dir, $_[0]);
+        next if -d $abs;
         return $abs if (-x $abs or $abs = is_executable($abs));
     }
 
@@ -1014,6 +1017,40 @@ sub is_path_inside {
     return substr ($path, 0, length ($parent)+1) eq "$parent/";
 }
 
+=head3 uri_escape($uri)
+
+Returns escaped URI.
+
+=cut
+
+sub uri_escape {
+    my ($uri) = @_;
+    $uri =~ s/([^0-9A-Za-z@%+\-\/:_.!~*'()])/sprintf("%%%02X", ord($1))/eg;
+    return $uri;
+}
+
+=head3 uri_unescape($uri)
+
+Unescape escaped URI and return it.
+
+=cut
+
+sub uri_unescape {
+    my ($uri) = @_;
+    $uri =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
+    return $uri;
+}
+
+=head3 is_depotpath($path)
+
+Check if a string is a valid depotpath.
+
+=cut
+
+sub is_depotpath {
+    ($_[0] =~ m|^/([^/]*)(/.*?)/?$|)
+}
+
 1;
 
 __END__
